# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import List, Union, Optional
from datetime import datetime
from typing_extensions import Annotated, TypeAlias

from ..._utils import PropertyInfo
from ..._models import BaseModel
from .plan_phase_maximum_adjustment import PlanPhaseMaximumAdjustment
from .plan_phase_minimum_adjustment import PlanPhaseMinimumAdjustment
from .plan_phase_usage_discount_adjustment import PlanPhaseUsageDiscountAdjustment
from .plan_phase_amount_discount_adjustment import PlanPhaseAmountDiscountAdjustment
from .plan_phase_percentage_discount_adjustment import PlanPhasePercentageDiscountAdjustment

__all__ = ["AdjustmentInterval", "Adjustment"]

Adjustment: TypeAlias = Annotated[
    Union[
        PlanPhaseUsageDiscountAdjustment,
        PlanPhaseAmountDiscountAdjustment,
        PlanPhasePercentageDiscountAdjustment,
        PlanPhaseMinimumAdjustment,
        PlanPhaseMaximumAdjustment,
    ],
    PropertyInfo(discriminator="adjustment_type"),
]


class AdjustmentInterval(BaseModel):
    id: str

    adjustment: Adjustment

    applies_to_price_interval_ids: List[str]
    """The price interval IDs that this adjustment applies to."""

    end_date: Optional[datetime] = None
    """The end date of the adjustment interval."""

    start_date: datetime
    """The start date of the adjustment interval."""



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import List, Optional

from ..._models import BaseModel
from .matrix_value import MatrixValue

__all__ = ["MatrixConfig"]


class MatrixConfig(BaseModel):
    default_unit_amount: str
    """Default per unit rate for any usage not bucketed into a specified matrix_value"""

    dimensions: List[Optional[str]]
    """One or two event property values to evaluate matrix groups by"""

    matrix_values: List[MatrixValue]
    """Matrix values configuration"""



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import Optional

from ..._models import BaseModel

__all__ = ["Tier"]


class Tier(BaseModel):
    first_unit: float
    """Exclusive tier starting value"""

    unit_amount: str
    """Amount per unit"""

    last_unit: Optional[float] = None
    """Inclusive tier ending value.

    This value is null if and only if this is the last tier.
    """



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from .tier import Tier as Tier
from .price import Price as Price
from .address import Address as Address
from .invoice import Invoice as Invoice
from .maximum import Maximum as Maximum
from .minimum import Minimum as Minimum
from .discount import Discount as Discount
from .bulk_tier import BulkTier as BulkTier
from .item_slim import ItemSlim as ItemSlim
from .allocation import Allocation as Allocation
from .tax_amount import TaxAmount as TaxAmount
from .bulk_config import BulkConfig as BulkConfig
from .credit_note import CreditNote as CreditNote
from .new_maximum import NewMaximum as NewMaximum
from .new_minimum import NewMinimum as NewMinimum
from .unit_config import UnitConfig as UnitConfig
from .invoice_tiny import InvoiceTiny as InvoiceTiny
from .matrix_value import MatrixValue as MatrixValue
from .matrix_config import MatrixConfig as MatrixConfig
from .tiered_config import TieredConfig as TieredConfig
from .package_config import PackageConfig as PackageConfig
from .per_price_cost import PerPriceCost as PerPriceCost
from .price_interval import PriceInterval as PriceInterval
from .trial_discount import TrialDiscount as TrialDiscount
from .usage_discount import UsageDiscount as UsageDiscount
from .aggregated_cost import AggregatedCost as AggregatedCost
from .amount_discount import AmountDiscount as AmountDiscount
from .customer_tax_id import CustomerTaxID as CustomerTaxID
from .credit_note_tiny import CreditNoteTiny as CreditNoteTiny
from .maximum_interval import MaximumInterval as MaximumInterval
from .minimum_interval import MinimumInterval as MinimumInterval
from .coupon_redemption import CouponRedemption as CouponRedemption
from .custom_expiration import CustomExpiration as CustomExpiration
from .customer_minified import CustomerMinified as CustomerMinified
from .new_usage_discount import NewUsageDiscount as NewUsageDiscount
from .tier_sub_line_item import TierSubLineItem as TierSubLineItem
from .adjustment_interval import AdjustmentInterval as AdjustmentInterval
from .new_amount_discount import NewAmountDiscount as NewAmountDiscount
from .new_plan_bulk_price import NewPlanBulkPrice as NewPlanBulkPrice
from .new_plan_unit_price import NewPlanUnitPrice as NewPlanUnitPrice
from .other_sub_line_item import OtherSubLineItem as OtherSubLineItem
from .pagination_metadata import PaginationMetadata as PaginationMetadata
from .percentage_discount import PercentageDiscount as PercentageDiscount
from .billable_metric_tiny import BillableMetricTiny as BillableMetricTiny
from .conversion_rate_tier import ConversionRateTier as ConversionRateTier
from .matrix_sub_line_item import MatrixSubLineItem as MatrixSubLineItem
from .new_allocation_price import NewAllocationPrice as NewAllocationPrice
from .new_plan_matrix_price import NewPlanMatrixPrice as NewPlanMatrixPrice
from .new_plan_tiered_price import NewPlanTieredPrice as NewPlanTieredPrice
from .subscription_minified import SubscriptionMinified as SubscriptionMinified
from .invoice_level_discount import InvoiceLevelDiscount as InvoiceLevelDiscount
from .new_plan_package_price import NewPlanPackagePrice as NewPlanPackagePrice
from .sub_line_item_grouping import SubLineItemGrouping as SubLineItemGrouping
from .transform_price_filter import TransformPriceFilter as TransformPriceFilter
from .new_floating_bulk_price import NewFloatingBulkPrice as NewFloatingBulkPrice
from .new_floating_unit_price import NewFloatingUnitPrice as NewFloatingUnitPrice
from .new_percentage_discount import NewPercentageDiscount as NewPercentageDiscount
from .subscription_trial_info import SubscriptionTrialInfo as SubscriptionTrialInfo
from .usage_discount_interval import UsageDiscountInterval as UsageDiscountInterval
from .amount_discount_interval import AmountDiscountInterval as AmountDiscountInterval
from .new_floating_matrix_price import NewFloatingMatrixPrice as NewFloatingMatrixPrice
from .new_floating_tiered_price import NewFloatingTieredPrice as NewFloatingTieredPrice
from .new_floating_package_price import NewFloatingPackagePrice as NewFloatingPackagePrice
from .billing_cycle_configuration import BillingCycleConfiguration as BillingCycleConfiguration
from .billing_cycle_relative_date import BillingCycleRelativeDate as BillingCycleRelativeDate
from .conversion_rate_unit_config import ConversionRateUnitConfig as ConversionRateUnitConfig
from .monetary_maximum_adjustment import MonetaryMaximumAdjustment as MonetaryMaximumAdjustment
from .monetary_minimum_adjustment import MonetaryMinimumAdjustment as MonetaryMinimumAdjustment
from .sub_line_item_matrix_config import SubLineItemMatrixConfig as SubLineItemMatrixConfig
from .unit_conversion_rate_config import UnitConversionRateConfig as UnitConversionRateConfig
from .percentage_discount_interval import PercentageDiscountInterval as PercentageDiscountInterval
from .subscription_change_minified import SubscriptionChangeMinified as SubscriptionChangeMinified
from .conversion_rate_tiered_config import ConversionRateTieredConfig as ConversionRateTieredConfig
from .fixed_fee_quantity_transition import FixedFeeQuantityTransition as FixedFeeQuantityTransition
from .matrix_with_allocation_config import MatrixWithAllocationConfig as MatrixWithAllocationConfig
from .new_plan_grouped_tiered_price import NewPlanGroupedTieredPrice as NewPlanGroupedTieredPrice
from .new_plan_tiered_package_price import NewPlanTieredPackagePrice as NewPlanTieredPackagePrice
from .plan_phase_maximum_adjustment import PlanPhaseMaximumAdjustment as PlanPhaseMaximumAdjustment
from .plan_phase_minimum_adjustment import PlanPhaseMinimumAdjustment as PlanPhaseMinimumAdjustment
from .tiered_conversion_rate_config import TieredConversionRateConfig as TieredConversionRateConfig
from .changed_subscription_resources import ChangedSubscriptionResources as ChangedSubscriptionResources
from .dimensional_price_configuration import DimensionalPriceConfiguration as DimensionalPriceConfiguration
from .new_billing_cycle_configuration import NewBillingCycleConfiguration as NewBillingCycleConfiguration
from .new_plan_minimum_composite_price import NewPlanMinimumCompositePrice as NewPlanMinimumCompositePrice
from .new_plan_unit_with_percent_price import NewPlanUnitWithPercentPrice as NewPlanUnitWithPercentPrice
from .fixed_fee_quantity_schedule_entry import FixedFeeQuantityScheduleEntry as FixedFeeQuantityScheduleEntry
from .new_floating_grouped_tiered_price import NewFloatingGroupedTieredPrice as NewFloatingGroupedTieredPrice
from .new_floating_tiered_package_price import NewFloatingTieredPackagePrice as NewFloatingTieredPackagePrice
from .new_plan_grouped_allocation_price import NewPlanGroupedAllocationPrice as NewPlanGroupedAllocationPrice
from .billing_cycle_anchor_configuration import BillingCycleAnchorConfiguration as BillingCycleAnchorConfiguration
from .monetary_usage_discount_adjustment import MonetaryUsageDiscountAdjustment as MonetaryUsageDiscountAdjustment
from .new_plan_bulk_with_proration_price import NewPlanBulkWithProrationPrice as NewPlanBulkWithProrationPrice
from .new_plan_tiered_with_minimum_price import NewPlanTieredWithMinimumPrice as NewPlanTieredWithMinimumPrice
from .new_plan_unit_with_proration_price import NewPlanUnitWithProrationPrice as NewPlanUnitWithProrationPrice
from .monetary_amount_discount_adjustment import MonetaryAmountDiscountAdjustment as MonetaryAmountDiscountAdjustment
from .new_dimensional_price_configuration import NewDimensionalPriceConfiguration as NewDimensionalPriceConfiguration
from .new_floating_minimum_composite_price import NewFloatingMinimumCompositePrice as NewFloatingMinimumCompositePrice
from .new_floating_unit_with_percent_price import NewFloatingUnitWithPercentPrice as NewFloatingUnitWithPercentPrice
from .plan_phase_usage_discount_adjustment import PlanPhaseUsageDiscountAdjustment as PlanPhaseUsageDiscountAdjustment
from .new_floating_grouped_allocation_price import (
    NewFloatingGroupedAllocationPrice as NewFloatingGroupedAllocationPrice,
)
from .new_plan_grouped_tiered_package_price import NewPlanGroupedTieredPackagePrice as NewPlanGroupedTieredPackagePrice
from .new_plan_matrix_with_allocation_price import NewPlanMatrixWithAllocationPrice as NewPlanMatrixWithAllocationPrice
from .new_plan_threshold_total_amount_price import NewPlanThresholdTotalAmountPrice as NewPlanThresholdTotalAmountPrice
from .plan_phase_amount_discount_adjustment import (
    PlanPhaseAmountDiscountAdjustment as PlanPhaseAmountDiscountAdjustment,
)
from .new_floating_bulk_with_proration_price import (
    NewFloatingBulkWithProrationPrice as NewFloatingBulkWithProrationPrice,
)
from .new_floating_tiered_with_minimum_price import (
    NewFloatingTieredWithMinimumPrice as NewFloatingTieredWithMinimumPrice,
)
from .new_floating_unit_with_proration_price import (
    NewFloatingUnitWithProrationPrice as NewFloatingUnitWithProrationPrice,
)
from .new_plan_cumulative_grouped_bulk_price import (
    NewPlanCumulativeGroupedBulkPrice as NewPlanCumulativeGroupedBulkPrice,
)
from .new_plan_package_with_allocation_price import (
    NewPlanPackageWithAllocationPrice as NewPlanPackageWithAllocationPrice,
)
from .monetary_percentage_discount_adjustment import (
    MonetaryPercentageDiscountAdjustment as MonetaryPercentageDiscountAdjustment,
)
from .new_plan_matrix_with_display_name_price import (
    NewPlanMatrixWithDisplayNamePrice as NewPlanMatrixWithDisplayNamePrice,
)
from .new_plan_max_group_tiered_package_price import (
    NewPlanMaxGroupTieredPackagePrice as NewPlanMaxGroupTieredPackagePrice,
)
from .new_floating_tiered_with_proration_price import (
    NewFloatingTieredWithProrationPrice as NewFloatingTieredWithProrationPrice,
)
from .new_floating_grouped_tiered_package_price import (
    NewFloatingGroupedTieredPackagePrice as NewFloatingGroupedTieredPackagePrice,
)
from .new_floating_matrix_with_allocation_price import (
    NewFloatingMatrixWithAllocationPrice as NewFloatingMatrixWithAllocationPrice,
)
from .new_floating_threshold_total_amount_price import (
    NewFloatingThresholdTotalAmountPrice as NewFloatingThresholdTotalAmountPrice,
)
from .plan_phase_percentage_discount_adjustment import (
    PlanPhasePercentageDiscountAdjustment as PlanPhasePercentageDiscountAdjustment,
)
from .new_floating_cumulative_grouped_bulk_price import (
    NewFloatingCumulativeGroupedBulkPrice as NewFloatingCumulativeGroupedBulkPrice,
)
from .new_floating_package_with_allocation_price import (
    NewFloatingPackageWithAllocationPrice as NewFloatingPackageWithAllocationPrice,
)
from .new_plan_tiered_package_with_minimum_price import (
    NewPlanTieredPackageWithMinimumPrice as NewPlanTieredPackageWithMinimumPrice,
)
from .new_floating_matrix_with_display_name_price import (
    NewFloatingMatrixWithDisplayNamePrice as NewFloatingMatrixWithDisplayNamePrice,
)
from .new_floating_max_group_tiered_package_price import (
    NewFloatingMaxGroupTieredPackagePrice as NewFloatingMaxGroupTieredPackagePrice,
)
from .new_plan_grouped_with_metered_minimum_price import (
    NewPlanGroupedWithMeteredMinimumPrice as NewPlanGroupedWithMeteredMinimumPrice,
)
from .new_plan_grouped_with_prorated_minimum_price import (
    NewPlanGroupedWithProratedMinimumPrice as NewPlanGroupedWithProratedMinimumPrice,
)
from .new_floating_tiered_package_with_minimum_price import (
    NewFloatingTieredPackageWithMinimumPrice as NewFloatingTieredPackageWithMinimumPrice,
)
from .new_floating_grouped_with_metered_minimum_price import (
    NewFloatingGroupedWithMeteredMinimumPrice as NewFloatingGroupedWithMeteredMinimumPrice,
)
from .new_floating_grouped_with_prorated_minimum_price import (
    NewFloatingGroupedWithProratedMinimumPrice as NewFloatingGroupedWithProratedMinimumPrice,
)
from .new_plan_scalable_matrix_with_unit_pricing_price import (
    NewPlanScalableMatrixWithUnitPricingPrice as NewPlanScalableMatrixWithUnitPricingPrice,
)
from .new_plan_scalable_matrix_with_tiered_pricing_price import (
    NewPlanScalableMatrixWithTieredPricingPrice as NewPlanScalableMatrixWithTieredPricingPrice,
)
from .new_floating_scalable_matrix_with_unit_pricing_price import (
    NewFloatingScalableMatrixWithUnitPricingPrice as NewFloatingScalableMatrixWithUnitPricingPrice,
)
from .new_floating_scalable_matrix_with_tiered_pricing_price import (
    NewFloatingScalableMatrixWithTieredPricingPrice as NewFloatingScalableMatrixWithTieredPricingPrice,
)



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import Optional

from ..._models import BaseModel

__all__ = ["BulkTier"]


class BulkTier(BaseModel):
    unit_amount: str
    """Amount per unit"""

    maximum_units: Optional[float] = None
    """Upper bound for this tier"""



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import Iterable, Optional
from typing_extensions import TypedDict

from .accounting_provider_config_param import AccountingProviderConfigParam

__all__ = ["NewAccountingSyncConfigurationParam"]


class NewAccountingSyncConfigurationParam(TypedDict, total=False):
    accounting_providers: Optional[Iterable[AccountingProviderConfigParam]]

    excluded: Optional[bool]



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import Optional
from typing_extensions import Literal

from ..._models import BaseModel
from .sub_line_item_grouping import SubLineItemGrouping

__all__ = ["OtherSubLineItem"]


class OtherSubLineItem(BaseModel):
    amount: str
    """The total amount for this sub line item."""

    grouping: Optional[SubLineItemGrouping] = None

    name: str

    quantity: float

    type: Literal["'null'"]



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

import os
from typing import Any, cast

import pytest

from orb import Orb, AsyncOrb
from orb._utils import parse_datetime
from tests.utils import assert_matches_type
from orb.types.customers import (
    CostListResponse,
    CostListByExternalIDResponse,
)

base_url = os.environ.get("TEST_API_BASE_URL", "http://127.0.0.1:4010")


class TestCosts:
    parametrize = pytest.mark.parametrize("client", [False, True], indirect=True, ids=["loose", "strict"])

    @parametrize
    def test_method_list(self, client: Orb) -> None:
        cost = client.customers.costs.list(
            customer_id="customer_id",
        )
        assert_matches_type(CostListResponse, cost, path=["response"])

    @parametrize
    def test_method_list_with_all_params(self, client: Orb) -> None:
        cost = client.customers.costs.list(
            customer_id="customer_id",
            currency="currency",
            timeframe_end=parse_datetime("2022-03-01T05:00:00Z"),
            timeframe_start=parse_datetime("2022-02-01T05:00:00Z"),
            view_mode="periodic",
        )
        assert_matches_type(CostListResponse, cost, path=["response"])

    @parametrize
    def test_raw_response_list(self, client: Orb) -> None:
        response = client.customers.costs.with_raw_response.list(
            customer_id="customer_id",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        cost = response.parse()
        assert_matches_type(CostListResponse, cost, path=["response"])

    @parametrize
    def test_streaming_response_list(self, client: Orb) -> None:
        with client.customers.costs.with_streaming_response.list(
            customer_id="customer_id",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            cost = response.parse()
            assert_matches_type(CostListResponse, cost, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    def test_path_params_list(self, client: Orb) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `customer_id` but received ''"):
            client.customers.costs.with_raw_response.list(
                customer_id="",
            )

    @parametrize
    def test_method_list_by_external_id(self, client: Orb) -> None:
        cost = client.customers.costs.list_by_external_id(
            external_customer_id="external_customer_id",
        )
        assert_matches_type(CostListByExternalIDResponse, cost, path=["response"])

    @parametrize
    def test_method_list_by_external_id_with_all_params(self, client: Orb) -> None:
        cost = client.customers.costs.list_by_external_id(
            external_customer_id="external_customer_id",
            currency="currency",
            timeframe_end=parse_datetime("2022-03-01T05:00:00Z"),
            timeframe_start=parse_datetime("2022-02-01T05:00:00Z"),
            view_mode="periodic",
        )
        assert_matches_type(CostListByExternalIDResponse, cost, path=["response"])

    @parametrize
    def test_raw_response_list_by_external_id(self, client: Orb) -> None:
        response = client.customers.costs.with_raw_response.list_by_external_id(
            external_customer_id="external_customer_id",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        cost = response.parse()
        assert_matches_type(CostListByExternalIDResponse, cost, path=["response"])

    @parametrize
    def test_streaming_response_list_by_external_id(self, client: Orb) -> None:
        with client.customers.costs.with_streaming_response.list_by_external_id(
            external_customer_id="external_customer_id",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            cost = response.parse()
            assert_matches_type(CostListByExternalIDResponse, cost, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    def test_path_params_list_by_external_id(self, client: Orb) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `external_customer_id` but received ''"):
            client.customers.costs.with_raw_response.list_by_external_id(
                external_customer_id="",
            )


class TestAsyncCosts:
    parametrize = pytest.mark.parametrize(
        "async_client", [False, True, {"http_client": "aiohttp"}], indirect=True, ids=["loose", "strict", "aiohttp"]
    )

    @parametrize
    async def test_method_list(self, async_client: AsyncOrb) -> None:
        cost = await async_client.customers.costs.list(
            customer_id="customer_id",
        )
        assert_matches_type(CostListResponse, cost, path=["response"])

    @parametrize
    async def test_method_list_with_all_params(self, async_client: AsyncOrb) -> None:
        cost = await async_client.customers.costs.list(
            customer_id="customer_id",
            currency="currency",
            timeframe_end=parse_datetime("2022-03-01T05:00:00Z"),
            timeframe_start=parse_datetime("2022-02-01T05:00:00Z"),
            view_mode="periodic",
        )
        assert_matches_type(CostListResponse, cost, path=["response"])

    @parametrize
    async def test_raw_response_list(self, async_client: AsyncOrb) -> None:
        response = await async_client.customers.costs.with_raw_response.list(
            customer_id="customer_id",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        cost = response.parse()
        assert_matches_type(CostListResponse, cost, path=["response"])

    @parametrize
    async def test_streaming_response_list(self, async_client: AsyncOrb) -> None:
        async with async_client.customers.costs.with_streaming_response.list(
            customer_id="customer_id",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            cost = await response.parse()
            assert_matches_type(CostListResponse, cost, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    async def test_path_params_list(self, async_client: AsyncOrb) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `customer_id` but received ''"):
            await async_client.customers.costs.with_raw_response.list(
                customer_id="",
            )

    @parametrize
    async def test_method_list_by_external_id(self, async_client: AsyncOrb) -> None:
        cost = await async_client.customers.costs.list_by_external_id(
            external_customer_id="external_customer_id",
        )
        assert_matches_type(CostListByExternalIDResponse, cost, path=["response"])

    @parametrize
    async def test_method_list_by_external_id_with_all_params(self, async_client: AsyncOrb) -> None:
        cost = await async_client.customers.costs.list_by_external_id(
            external_customer_id="external_customer_id",
            currency="currency",
            timeframe_end=parse_datetime("2022-03-01T05:00:00Z"),
            timeframe_start=parse_datetime("2022-02-01T05:00:00Z"),
            view_mode="periodic",
        )
        assert_matches_type(CostListByExternalIDResponse, cost, path=["response"])

    @parametrize
    async def test_raw_response_list_by_external_id(self, async_client: AsyncOrb) -> None:
        response = await async_client.customers.costs.with_raw_response.list_by_external_id(
            external_customer_id="external_customer_id",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        cost = response.parse()
        assert_matches_type(CostListByExternalIDResponse, cost, path=["response"])

    @parametrize
    async def test_streaming_response_list_by_external_id(self, async_client: AsyncOrb) -> None:
        async with async_client.customers.costs.with_streaming_response.list_by_external_id(
            external_customer_id="external_customer_id",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            cost = await response.parse()
            assert_matches_type(CostListByExternalIDResponse, cost, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    async def test_path_params_list_by_external_id(self, async_client: AsyncOrb) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `external_customer_id` but received ''"):
            await async_client.customers.costs.with_raw_response.list_by_external_id(
                external_customer_id="",
            )



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import Dict, List
from datetime import datetime
from typing_extensions import Literal

from .._models import BaseModel

__all__ = ["Item", "ExternalConnection"]


class ExternalConnection(BaseModel):
    external_connection_name: Literal[
        "stripe", "quickbooks", "bill.com", "netsuite", "taxjar", "avalara", "anrok", "numeral"
    ]

    external_entity_id: str


class Item(BaseModel):
    id: str

    created_at: datetime

    external_connections: List[ExternalConnection]

    metadata: Dict[str, str]
    """User specified key-value pairs for the resource.

    If not present, this defaults to an empty dictionary. Individual keys can be
    removed by setting the value to `null`, and the entire metadata mapping can be
    cleared by setting `metadata` to `null`.
    """

    name: str



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import List, Optional
from datetime import datetime
from typing_extensions import Literal

from .._models import BaseModel
from .threshold import Threshold
from .shared.customer_minified import CustomerMinified
from .shared.subscription_minified import SubscriptionMinified

__all__ = ["Alert", "Metric", "Plan", "BalanceAlertStatus"]


class Metric(BaseModel):
    id: str


class Plan(BaseModel):
    id: Optional[str] = None

    external_plan_id: Optional[str] = None
    """
    An optional user-defined ID for this plan resource, used throughout the system
    as an alias for this Plan. Use this field to identify a plan by an existing
    identifier in your system.
    """

    name: Optional[str] = None

    plan_version: str


class BalanceAlertStatus(BaseModel):
    in_alert: bool
    """Whether the alert is currently in-alert or not."""

    threshold_value: float
    """The value of the threshold that defines the alert status."""


class Alert(BaseModel):
    id: str
    """Also referred to as alert_id in this documentation."""

    created_at: datetime
    """The creation time of the resource in Orb."""

    currency: Optional[str] = None
    """The name of the currency the credit balance or invoice cost is denominated in."""

    customer: Optional[CustomerMinified] = None
    """The customer the alert applies to."""

    enabled: bool
    """Whether the alert is enabled or disabled."""

    metric: Optional[Metric] = None
    """The metric the alert applies to."""

    plan: Optional[Plan] = None
    """The plan the alert applies to."""

    subscription: Optional[SubscriptionMinified] = None
    """The subscription the alert applies to."""

    thresholds: Optional[List[Threshold]] = None
    """
    The thresholds that define the conditions under which the alert will be
    triggered.
    """

    type: Literal[
        "credit_balance_depleted",
        "credit_balance_dropped",
        "credit_balance_recovered",
        "usage_exceeded",
        "cost_exceeded",
    ]
    """The type of alert. This must be a valid alert type."""

    balance_alert_status: Optional[List[BalanceAlertStatus]] = None
    """The current status of the alert.

    This field is only present for credit balance alerts.
    """



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import Union
from datetime import datetime
from typing_extensions import Literal, Required, Annotated, TypedDict

from .._utils import PropertyInfo

__all__ = ["SubscriptionUpdateTrialParams"]


class SubscriptionUpdateTrialParams(TypedDict, total=False):
    trial_end_date: Required[
        Annotated[Union[Union[str, datetime], Literal["immediate"]], PropertyInfo(format="iso8601")]
    ]
    """
    The new date that the trial should end, or the literal string `immediate` to end
    the trial immediately.
    """

    shift: bool
    """
    If true, shifts subsequent price and adjustment intervals (preserving their
    durations, but adjusting their absolute dates).
    """



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import Union, Optional
from datetime import datetime
from typing_extensions import Annotated, TypeAlias

from .._utils import PropertyInfo
from .._models import BaseModel
from .shared.amount_discount import AmountDiscount
from .shared.percentage_discount import PercentageDiscount

__all__ = ["Coupon", "Discount"]

Discount: TypeAlias = Annotated[Union[PercentageDiscount, AmountDiscount], PropertyInfo(discriminator="discount_type")]


class Coupon(BaseModel):
    id: str
    """Also referred to as coupon_id in this documentation."""

    archived_at: Optional[datetime] = None
    """An archived coupon can no longer be redeemed.

    Active coupons will have a value of null for `archived_at`; this field will be
    non-null for archived coupons.
    """

    discount: Discount

    duration_in_months: Optional[int] = None
    """
    This allows for a coupon's discount to apply for a limited time (determined in
    months); a `null` value here means "unlimited time".
    """

    max_redemptions: Optional[int] = None
    """
    The maximum number of redemptions allowed for this coupon before it is
    exhausted; `null` here means "unlimited".
    """

    redemption_code: str
    """This string can be used to redeem this coupon for a given subscription."""

    times_redeemed: int
    """The number of times this coupon has been redeemed."""



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import List, Union, Optional
from datetime import datetime
from typing_extensions import Annotated, TypeAlias

from .._utils import PropertyInfo
from .._models import BaseModel
from .shared.price import Price
from .shared.maximum import Maximum
from .shared.minimum import Minimum
from .shared.discount import Discount
from .shared.tax_amount import TaxAmount
from .shared.tier_sub_line_item import TierSubLineItem
from .shared.other_sub_line_item import OtherSubLineItem
from .shared.matrix_sub_line_item import MatrixSubLineItem
from .shared.monetary_maximum_adjustment import MonetaryMaximumAdjustment
from .shared.monetary_minimum_adjustment import MonetaryMinimumAdjustment
from .shared.monetary_usage_discount_adjustment import MonetaryUsageDiscountAdjustment
from .shared.monetary_amount_discount_adjustment import MonetaryAmountDiscountAdjustment
from .shared.monetary_percentage_discount_adjustment import MonetaryPercentageDiscountAdjustment

__all__ = ["InvoiceLineItemCreateResponse", "Adjustment", "SubLineItem"]

Adjustment: TypeAlias = Annotated[
    Union[
        MonetaryUsageDiscountAdjustment,
        MonetaryAmountDiscountAdjustment,
        MonetaryPercentageDiscountAdjustment,
        MonetaryMinimumAdjustment,
        MonetaryMaximumAdjustment,
    ],
    PropertyInfo(discriminator="adjustment_type"),
]

SubLineItem: TypeAlias = Annotated[
    Union[MatrixSubLineItem, TierSubLineItem, OtherSubLineItem], PropertyInfo(discriminator="type")
]


class InvoiceLineItemCreateResponse(BaseModel):
    id: str
    """A unique ID for this line item."""

    adjusted_subtotal: str
    """
    The line amount after any adjustments and before overage conversion, credits and
    partial invoicing.
    """

    adjustments: List[Adjustment]
    """
    All adjustments applied to the line item in the order they were applied based on
    invoice calculations (ie. usage discounts -> amount discounts -> percentage
    discounts -> minimums -> maximums).
    """

    amount: str
    """
    The final amount for a line item after all adjustments and pre paid credits have
    been applied.
    """

    credits_applied: str
    """The number of prepaid credits applied."""

    discount: Optional[Discount] = None
    """This field is deprecated in favor of `adjustments`"""

    end_date: datetime
    """The end date of the range of time applied for this line item's price."""

    filter: Optional[str] = None
    """An additional filter that was used to calculate the usage for this line item."""

    grouping: Optional[str] = None
    """
    [DEPRECATED] For configured prices that are split by a grouping key, this will
    be populated with the key and a value. The `amount` and `subtotal` will be the
    values for this particular grouping.
    """

    maximum: Optional[Maximum] = None
    """This field is deprecated in favor of `adjustments`."""

    maximum_amount: Optional[str] = None
    """This field is deprecated in favor of `adjustments`."""

    minimum: Optional[Minimum] = None
    """This field is deprecated in favor of `adjustments`."""

    minimum_amount: Optional[str] = None
    """This field is deprecated in favor of `adjustments`."""

    name: str
    """The name of the price associated with this line item."""

    partially_invoiced_amount: str
    """Any amount applied from a partial invoice"""

    price: Price
    """
    The Price resource represents a price that can be billed on a subscription,
    resulting in a charge on an invoice in the form of an invoice line item. Prices
    take a quantity and determine an amount to bill.

    Orb supports a few different pricing models out of the box. Each of these models
    is serialized differently in a given Price object. The model_type field
    determines the key for the configuration object that is present.

    For more on the types of prices, see
    [the core concepts documentation](/core-concepts#plan-and-price)
    """

    quantity: float
    """Either the fixed fee quantity or the usage during the service period."""

    start_date: datetime
    """The start date of the range of time applied for this line item's price."""

    sub_line_items: List[SubLineItem]
    """
    For complex pricing structures, the line item can be broken down further in
    `sub_line_items`.
    """

    subtotal: str
    """The line amount before any adjustments."""

    tax_amounts: List[TaxAmount]
    """An array of tax rates and their incurred tax amounts.

    Empty if no tax integration is configured.
    """

    usage_customer_ids: Optional[List[str]] = None
    """
    A list of customer ids that were used to calculate the usage for this line item.
    """



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import Dict, Optional
from typing_extensions import TypedDict

__all__ = ["DimensionalPriceGroupUpdateParams"]


class DimensionalPriceGroupUpdateParams(TypedDict, total=False):
    external_dimensional_price_group_id: Optional[str]
    """
    An optional user-defined ID for this dimensional price group resource, used
    throughout the system as an alias for this dimensional price group. Use this
    field to identify a dimensional price group by an existing identifier in your
    system.
    """

    metadata: Optional[Dict[str, Optional[str]]]
    """User-specified key/value pairs for the resource.

    Individual keys can be removed by setting the value to `null`, and the entire
    metadata mapping can be cleared by setting `metadata` to `null`.
    """



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing_extensions import Literal

from ..._models import BaseModel

__all__ = ["BillingCycleConfiguration"]


class BillingCycleConfiguration(BaseModel):
    duration: int

    duration_unit: Literal["day", "month"]



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import Optional
from typing_extensions import TypedDict

__all__ = ["CreditListByExternalIDParams"]


class CreditListByExternalIDParams(TypedDict, total=False):
    currency: Optional[str]
    """The ledger currency or custom pricing unit to use."""

    cursor: Optional[str]
    """Cursor for pagination.

    This can be populated by the `next_cursor` value returned from the initial
    request.
    """

    include_all_blocks: bool
    """
    If set to True, all expired and depleted blocks, as well as active block will be
    returned.
    """

    limit: int
    """The number of items to fetch. Defaults to 20."""



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from datetime import datetime

from ..._models import BaseModel

__all__ = ["FixedFeeQuantityTransition"]


class FixedFeeQuantityTransition(BaseModel):
    effective_date: datetime

    price_id: str

    quantity: int



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import Union
from typing_extensions import Annotated, TypeAlias

from ...._utils import PropertyInfo
from .void_ledger_entry import VoidLedgerEntry
from .amendment_ledger_entry import AmendmentLedgerEntry
from .decrement_ledger_entry import DecrementLedgerEntry
from .increment_ledger_entry import IncrementLedgerEntry
from .void_initiated_ledger_entry import VoidInitiatedLedgerEntry
from .expiration_change_ledger_entry import ExpirationChangeLedgerEntry
from .credit_block_expiry_ledger_entry import CreditBlockExpiryLedgerEntry

__all__ = ["LedgerListByExternalIDResponse"]

LedgerListByExternalIDResponse: TypeAlias = Annotated[
    Union[
        IncrementLedgerEntry,
        DecrementLedgerEntry,
        ExpirationChangeLedgerEntry,
        CreditBlockExpiryLedgerEntry,
        VoidLedgerEntry,
        VoidInitiatedLedgerEntry,
        AmendmentLedgerEntry,
    ],
    PropertyInfo(discriminator="entry_type"),
]



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing_extensions import Literal

from ..._models import BaseModel

__all__ = ["CustomerTaxID"]


class CustomerTaxID(BaseModel):
    country: Literal[
        "AD",
        "AE",
        "AL",
        "AM",
        "AO",
        "AR",
        "AT",
        "AU",
        "AW",
        "AZ",
        "BA",
        "BB",
        "BD",
        "BE",
        "BF",
        "BG",
        "BH",
        "BJ",
        "BO",
        "BR",
        "BS",
        "BY",
        "CA",
        "CD",
        "CH",
        "CL",
        "CM",
        "CN",
        "CO",
        "CR",
        "CV",
        "DE",
        "CY",
        "CZ",
        "DK",
        "DO",
        "EC",
        "EE",
        "EG",
        "ES",
        "ET",
        "EU",
        "FI",
        "FR",
        "GB",
        "GE",
        "GN",
        "GR",
        "HK",
        "HR",
        "HU",
        "ID",
        "IE",
        "IL",
        "IN",
        "IS",
        "IT",
        "JP",
        "KE",
        "KG",
        "KH",
        "KR",
        "KZ",
        "LA",
        "LI",
        "LT",
        "LU",
        "LV",
        "MA",
        "MD",
        "ME",
        "MK",
        "MR",
        "MT",
        "MX",
        "MY",
        "NG",
        "NL",
        "NO",
        "NP",
        "NZ",
        "OM",
        "PE",
        "PH",
        "PL",
        "PT",
        "RO",
        "RS",
        "RU",
        "SA",
        "SE",
        "SG",
        "SI",
        "SK",
        "SN",
        "SR",
        "SV",
        "TH",
        "TJ",
        "TR",
        "TW",
        "TZ",
        "UA",
        "UG",
        "US",
        "UY",
        "UZ",
        "VE",
        "VN",
        "ZA",
        "ZM",
        "ZW",
    ]

    type: Literal[
        "ad_nrt",
        "ae_trn",
        "al_tin",
        "am_tin",
        "ao_tin",
        "ar_cuit",
        "eu_vat",
        "au_abn",
        "au_arn",
        "aw_tin",
        "az_tin",
        "ba_tin",
        "bb_tin",
        "bd_bin",
        "bf_ifu",
        "bg_uic",
        "bh_vat",
        "bj_ifu",
        "bo_tin",
        "br_cnpj",
        "br_cpf",
        "bs_tin",
        "by_tin",
        "ca_bn",
        "ca_gst_hst",
        "ca_pst_bc",
        "ca_pst_mb",
        "ca_pst_sk",
        "ca_qst",
        "cd_nif",
        "ch_uid",
        "ch_vat",
        "cl_tin",
        "cm_niu",
        "cn_tin",
        "co_nit",
        "cr_tin",
        "cv_nif",
        "de_stn",
        "do_rcn",
        "ec_ruc",
        "eg_tin",
        "es_cif",
        "et_tin",
        "eu_oss_vat",
        "gb_vat",
        "ge_vat",
        "gn_nif",
        "hk_br",
        "hr_oib",
        "hu_tin",
        "id_npwp",
        "il_vat",
        "in_gst",
        "is_vat",
        "jp_cn",
        "jp_rn",
        "jp_trn",
        "ke_pin",
        "kg_tin",
        "kh_tin",
        "kr_brn",
        "kz_bin",
        "la_tin",
        "li_uid",
        "li_vat",
        "ma_vat",
        "md_vat",
        "me_pib",
        "mk_vat",
        "mr_nif",
        "mx_rfc",
        "my_frp",
        "my_itn",
        "my_sst",
        "ng_tin",
        "no_vat",
        "no_voec",
        "np_pan",
        "nz_gst",
        "om_vat",
        "pe_ruc",
        "ph_tin",
        "ro_tin",
        "rs_pib",
        "ru_inn",
        "ru_kpp",
        "sa_vat",
        "sg_gst",
        "sg_uen",
        "si_tin",
        "sn_ninea",
        "sr_fin",
        "sv_nit",
        "th_vat",
        "tj_tin",
        "tr_tin",
        "tw_vat",
        "tz_vat",
        "ua_vat",
        "ug_tin",
        "us_ein",
        "uy_ruc",
        "uz_tin",
        "uz_vat",
        "ve_rif",
        "vn_tin",
        "za_vat",
        "zm_tin",
        "zw_tin",
    ]

    value: str



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

import os
from typing import Any, cast

import pytest

from orb import Orb, AsyncOrb
from orb.types import (
    InvoiceFetchUpcomingResponse,
)
from orb._utils import parse_date, parse_datetime
from tests.utils import assert_matches_type
from orb.pagination import SyncPage, AsyncPage
from orb.types.shared import Invoice

base_url = os.environ.get("TEST_API_BASE_URL", "http://127.0.0.1:4010")


class TestInvoices:
    parametrize = pytest.mark.parametrize("client", [False, True], indirect=True, ids=["loose", "strict"])

    @parametrize
    def test_method_create(self, client: Orb) -> None:
        invoice = client.invoices.create(
            currency="USD",
            invoice_date=parse_datetime("2019-12-27T18:11:19.117Z"),
            line_items=[
                {
                    "end_date": parse_date("2023-09-22"),
                    "item_id": "4khy3nwzktxv7",
                    "model_type": "unit",
                    "name": "Line Item Name",
                    "quantity": 1,
                    "start_date": parse_date("2023-09-22"),
                    "unit_config": {"unit_amount": "unit_amount"},
                }
            ],
        )
        assert_matches_type(Invoice, invoice, path=["response"])

    @parametrize
    def test_method_create_with_all_params(self, client: Orb) -> None:
        invoice = client.invoices.create(
            currency="USD",
            invoice_date=parse_datetime("2019-12-27T18:11:19.117Z"),
            line_items=[
                {
                    "end_date": parse_date("2023-09-22"),
                    "item_id": "4khy3nwzktxv7",
                    "model_type": "unit",
                    "name": "Line Item Name",
                    "quantity": 1,
                    "start_date": parse_date("2023-09-22"),
                    "unit_config": {
                        "unit_amount": "unit_amount",
                        "scaling_factor": 0,
                    },
                }
            ],
            customer_id="4khy3nwzktxv7",
            discount={
                "discount_type": "percentage",
                "percentage_discount": 0.15,
                "applies_to_price_ids": ["h74gfhdjvn7ujokd", "7hfgtgjnbvc3ujkl"],
                "filters": [
                    {
                        "field": "price_id",
                        "operator": "includes",
                        "values": ["string"],
                    }
                ],
                "reason": "reason",
            },
            due_date=parse_date("2023-09-22"),
            external_customer_id="external-customer-id",
            memo="An optional memo for my invoice.",
            metadata={"foo": "string"},
            net_terms=0,
            will_auto_issue=False,
        )
        assert_matches_type(Invoice, invoice, path=["response"])

    @parametrize
    def test_raw_response_create(self, client: Orb) -> None:
        response = client.invoices.with_raw_response.create(
            currency="USD",
            invoice_date=parse_datetime("2019-12-27T18:11:19.117Z"),
            line_items=[
                {
                    "end_date": parse_date("2023-09-22"),
                    "item_id": "4khy3nwzktxv7",
                    "model_type": "unit",
                    "name": "Line Item Name",
                    "quantity": 1,
                    "start_date": parse_date("2023-09-22"),
                    "unit_config": {"unit_amount": "unit_amount"},
                }
            ],
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        invoice = response.parse()
        assert_matches_type(Invoice, invoice, path=["response"])

    @parametrize
    def test_streaming_response_create(self, client: Orb) -> None:
        with client.invoices.with_streaming_response.create(
            currency="USD",
            invoice_date=parse_datetime("2019-12-27T18:11:19.117Z"),
            line_items=[
                {
                    "end_date": parse_date("2023-09-22"),
                    "item_id": "4khy3nwzktxv7",
                    "model_type": "unit",
                    "name": "Line Item Name",
                    "quantity": 1,
                    "start_date": parse_date("2023-09-22"),
                    "unit_config": {"unit_amount": "unit_amount"},
                }
            ],
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            invoice = response.parse()
            assert_matches_type(Invoice, invoice, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    def test_method_update(self, client: Orb) -> None:
        invoice = client.invoices.update(
            invoice_id="invoice_id",
        )
        assert_matches_type(Invoice, invoice, path=["response"])

    @parametrize
    def test_method_update_with_all_params(self, client: Orb) -> None:
        invoice = client.invoices.update(
            invoice_id="invoice_id",
            due_date=parse_date("2023-09-22"),
            metadata={"foo": "string"},
            net_terms=0,
        )
        assert_matches_type(Invoice, invoice, path=["response"])

    @parametrize
    def test_raw_response_update(self, client: Orb) -> None:
        response = client.invoices.with_raw_response.update(
            invoice_id="invoice_id",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        invoice = response.parse()
        assert_matches_type(Invoice, invoice, path=["response"])

    @parametrize
    def test_streaming_response_update(self, client: Orb) -> None:
        with client.invoices.with_streaming_response.update(
            invoice_id="invoice_id",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            invoice = response.parse()
            assert_matches_type(Invoice, invoice, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    def test_path_params_update(self, client: Orb) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `invoice_id` but received ''"):
            client.invoices.with_raw_response.update(
                invoice_id="",
            )

    @parametrize
    def test_method_list(self, client: Orb) -> None:
        invoice = client.invoices.list()
        assert_matches_type(SyncPage[Invoice], invoice, path=["response"])

    @parametrize
    def test_method_list_with_all_params(self, client: Orb) -> None:
        invoice = client.invoices.list(
            amount="amount",
            amount_gt="amount[gt]",
            amount_lt="amount[lt]",
            cursor="cursor",
            customer_id="customer_id",
            date_type="due_date",
            due_date=parse_date("2019-12-27"),
            due_date_window="due_date_window",
            due_date_gt=parse_date("2019-12-27"),
            due_date_lt=parse_date("2019-12-27"),
            external_customer_id="external_customer_id",
            invoice_date_gt=parse_datetime("2019-12-27T18:11:19.117Z"),
            invoice_date_gte=parse_datetime("2019-12-27T18:11:19.117Z"),
            invoice_date_lt=parse_datetime("2019-12-27T18:11:19.117Z"),
            invoice_date_lte=parse_datetime("2019-12-27T18:11:19.117Z"),
            is_recurring=True,
            limit=1,
            status=["draft"],
            subscription_id="subscription_id",
        )
        assert_matches_type(SyncPage[Invoice], invoice, path=["response"])

    @parametrize
    def test_raw_response_list(self, client: Orb) -> None:
        response = client.invoices.with_raw_response.list()

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        invoice = response.parse()
        assert_matches_type(SyncPage[Invoice], invoice, path=["response"])

    @parametrize
    def test_streaming_response_list(self, client: Orb) -> None:
        with client.invoices.with_streaming_response.list() as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            invoice = response.parse()
            assert_matches_type(SyncPage[Invoice], invoice, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    def test_method_fetch(self, client: Orb) -> None:
        invoice = client.invoices.fetch(
            "invoice_id",
        )
        assert_matches_type(Invoice, invoice, path=["response"])

    @parametrize
    def test_raw_response_fetch(self, client: Orb) -> None:
        response = client.invoices.with_raw_response.fetch(
            "invoice_id",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        invoice = response.parse()
        assert_matches_type(Invoice, invoice, path=["response"])

    @parametrize
    def test_streaming_response_fetch(self, client: Orb) -> None:
        with client.invoices.with_streaming_response.fetch(
            "invoice_id",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            invoice = response.parse()
            assert_matches_type(Invoice, invoice, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    def test_path_params_fetch(self, client: Orb) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `invoice_id` but received ''"):
            client.invoices.with_raw_response.fetch(
                "",
            )

    @parametrize
    def test_method_fetch_upcoming(self, client: Orb) -> None:
        invoice = client.invoices.fetch_upcoming(
            subscription_id="subscription_id",
        )
        assert_matches_type(InvoiceFetchUpcomingResponse, invoice, path=["response"])

    @parametrize
    def test_raw_response_fetch_upcoming(self, client: Orb) -> None:
        response = client.invoices.with_raw_response.fetch_upcoming(
            subscription_id="subscription_id",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        invoice = response.parse()
        assert_matches_type(InvoiceFetchUpcomingResponse, invoice, path=["response"])

    @parametrize
    def test_streaming_response_fetch_upcoming(self, client: Orb) -> None:
        with client.invoices.with_streaming_response.fetch_upcoming(
            subscription_id="subscription_id",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            invoice = response.parse()
            assert_matches_type(InvoiceFetchUpcomingResponse, invoice, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    def test_method_issue(self, client: Orb) -> None:
        invoice = client.invoices.issue(
            invoice_id="invoice_id",
        )
        assert_matches_type(Invoice, invoice, path=["response"])

    @parametrize
    def test_method_issue_with_all_params(self, client: Orb) -> None:
        invoice = client.invoices.issue(
            invoice_id="invoice_id",
            synchronous=True,
        )
        assert_matches_type(Invoice, invoice, path=["response"])

    @parametrize
    def test_raw_response_issue(self, client: Orb) -> None:
        response = client.invoices.with_raw_response.issue(
            invoice_id="invoice_id",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        invoice = response.parse()
        assert_matches_type(Invoice, invoice, path=["response"])

    @parametrize
    def test_streaming_response_issue(self, client: Orb) -> None:
        with client.invoices.with_streaming_response.issue(
            invoice_id="invoice_id",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            invoice = response.parse()
            assert_matches_type(Invoice, invoice, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    def test_path_params_issue(self, client: Orb) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `invoice_id` but received ''"):
            client.invoices.with_raw_response.issue(
                invoice_id="",
            )

    @parametrize
    def test_method_mark_paid(self, client: Orb) -> None:
        invoice = client.invoices.mark_paid(
            invoice_id="invoice_id",
            payment_received_date=parse_date("2023-09-22"),
        )
        assert_matches_type(Invoice, invoice, path=["response"])

    @parametrize
    def test_method_mark_paid_with_all_params(self, client: Orb) -> None:
        invoice = client.invoices.mark_paid(
            invoice_id="invoice_id",
            payment_received_date=parse_date("2023-09-22"),
            external_id="external_payment_id_123",
            notes="notes",
        )
        assert_matches_type(Invoice, invoice, path=["response"])

    @parametrize
    def test_raw_response_mark_paid(self, client: Orb) -> None:
        response = client.invoices.with_raw_response.mark_paid(
            invoice_id="invoice_id",
            payment_received_date=parse_date("2023-09-22"),
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        invoice = response.parse()
        assert_matches_type(Invoice, invoice, path=["response"])

    @parametrize
    def test_streaming_response_mark_paid(self, client: Orb) -> None:
        with client.invoices.with_streaming_response.mark_paid(
            invoice_id="invoice_id",
            payment_received_date=parse_date("2023-09-22"),
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            invoice = response.parse()
            assert_matches_type(Invoice, invoice, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    def test_path_params_mark_paid(self, client: Orb) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `invoice_id` but received ''"):
            client.invoices.with_raw_response.mark_paid(
                invoice_id="",
                payment_received_date=parse_date("2023-09-22"),
            )

    @parametrize
    def test_method_pay(self, client: Orb) -> None:
        invoice = client.invoices.pay(
            "invoice_id",
        )
        assert_matches_type(Invoice, invoice, path=["response"])

    @parametrize
    def test_raw_response_pay(self, client: Orb) -> None:
        response = client.invoices.with_raw_response.pay(
            "invoice_id",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        invoice = response.parse()
        assert_matches_type(Invoice, invoice, path=["response"])

    @parametrize
    def test_streaming_response_pay(self, client: Orb) -> None:
        with client.invoices.with_streaming_response.pay(
            "invoice_id",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            invoice = response.parse()
            assert_matches_type(Invoice, invoice, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    def test_path_params_pay(self, client: Orb) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `invoice_id` but received ''"):
            client.invoices.with_raw_response.pay(
                "",
            )

    @parametrize
    def test_method_void(self, client: Orb) -> None:
        invoice = client.invoices.void(
            "invoice_id",
        )
        assert_matches_type(Invoice, invoice, path=["response"])

    @parametrize
    def test_raw_response_void(self, client: Orb) -> None:
        response = client.invoices.with_raw_response.void(
            "invoice_id",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        invoice = response.parse()
        assert_matches_type(Invoice, invoice, path=["response"])

    @parametrize
    def test_streaming_response_void(self, client: Orb) -> None:
        with client.invoices.with_streaming_response.void(
            "invoice_id",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            invoice = response.parse()
            assert_matches_type(Invoice, invoice, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    def test_path_params_void(self, client: Orb) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `invoice_id` but received ''"):
            client.invoices.with_raw_response.void(
                "",
            )


class TestAsyncInvoices:
    parametrize = pytest.mark.parametrize(
        "async_client", [False, True, {"http_client": "aiohttp"}], indirect=True, ids=["loose", "strict", "aiohttp"]
    )

    @parametrize
    async def test_method_create(self, async_client: AsyncOrb) -> None:
        invoice = await async_client.invoices.create(
            currency="USD",
            invoice_date=parse_datetime("2019-12-27T18:11:19.117Z"),
            line_items=[
                {
                    "end_date": parse_date("2023-09-22"),
                    "item_id": "4khy3nwzktxv7",
                    "model_type": "unit",
                    "name": "Line Item Name",
                    "quantity": 1,
                    "start_date": parse_date("2023-09-22"),
                    "unit_config": {"unit_amount": "unit_amount"},
                }
            ],
        )
        assert_matches_type(Invoice, invoice, path=["response"])

    @parametrize
    async def test_method_create_with_all_params(self, async_client: AsyncOrb) -> None:
        invoice = await async_client.invoices.create(
            currency="USD",
            invoice_date=parse_datetime("2019-12-27T18:11:19.117Z"),
            line_items=[
                {
                    "end_date": parse_date("2023-09-22"),
                    "item_id": "4khy3nwzktxv7",
                    "model_type": "unit",
                    "name": "Line Item Name",
                    "quantity": 1,
                    "start_date": parse_date("2023-09-22"),
                    "unit_config": {
                        "unit_amount": "unit_amount",
                        "scaling_factor": 0,
                    },
                }
            ],
            customer_id="4khy3nwzktxv7",
            discount={
                "discount_type": "percentage",
                "percentage_discount": 0.15,
                "applies_to_price_ids": ["h74gfhdjvn7ujokd", "7hfgtgjnbvc3ujkl"],
                "filters": [
                    {
                        "field": "price_id",
                        "operator": "includes",
                        "values": ["string"],
                    }
                ],
                "reason": "reason",
            },
            due_date=parse_date("2023-09-22"),
            external_customer_id="external-customer-id",
            memo="An optional memo for my invoice.",
            metadata={"foo": "string"},
            net_terms=0,
            will_auto_issue=False,
        )
        assert_matches_type(Invoice, invoice, path=["response"])

    @parametrize
    async def test_raw_response_create(self, async_client: AsyncOrb) -> None:
        response = await async_client.invoices.with_raw_response.create(
            currency="USD",
            invoice_date=parse_datetime("2019-12-27T18:11:19.117Z"),
            line_items=[
                {
                    "end_date": parse_date("2023-09-22"),
                    "item_id": "4khy3nwzktxv7",
                    "model_type": "unit",
                    "name": "Line Item Name",
                    "quantity": 1,
                    "start_date": parse_date("2023-09-22"),
                    "unit_config": {"unit_amount": "unit_amount"},
                }
            ],
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        invoice = response.parse()
        assert_matches_type(Invoice, invoice, path=["response"])

    @parametrize
    async def test_streaming_response_create(self, async_client: AsyncOrb) -> None:
        async with async_client.invoices.with_streaming_response.create(
            currency="USD",
            invoice_date=parse_datetime("2019-12-27T18:11:19.117Z"),
            line_items=[
                {
                    "end_date": parse_date("2023-09-22"),
                    "item_id": "4khy3nwzktxv7",
                    "model_type": "unit",
                    "name": "Line Item Name",
                    "quantity": 1,
                    "start_date": parse_date("2023-09-22"),
                    "unit_config": {"unit_amount": "unit_amount"},
                }
            ],
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            invoice = await response.parse()
            assert_matches_type(Invoice, invoice, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    async def test_method_update(self, async_client: AsyncOrb) -> None:
        invoice = await async_client.invoices.update(
            invoice_id="invoice_id",
        )
        assert_matches_type(Invoice, invoice, path=["response"])

    @parametrize
    async def test_method_update_with_all_params(self, async_client: AsyncOrb) -> None:
        invoice = await async_client.invoices.update(
            invoice_id="invoice_id",
            due_date=parse_date("2023-09-22"),
            metadata={"foo": "string"},
            net_terms=0,
        )
        assert_matches_type(Invoice, invoice, path=["response"])

    @parametrize
    async def test_raw_response_update(self, async_client: AsyncOrb) -> None:
        response = await async_client.invoices.with_raw_response.update(
            invoice_id="invoice_id",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        invoice = response.parse()
        assert_matches_type(Invoice, invoice, path=["response"])

    @parametrize
    async def test_streaming_response_update(self, async_client: AsyncOrb) -> None:
        async with async_client.invoices.with_streaming_response.update(
            invoice_id="invoice_id",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            invoice = await response.parse()
            assert_matches_type(Invoice, invoice, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    async def test_path_params_update(self, async_client: AsyncOrb) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `invoice_id` but received ''"):
            await async_client.invoices.with_raw_response.update(
                invoice_id="",
            )

    @parametrize
    async def test_method_list(self, async_client: AsyncOrb) -> None:
        invoice = await async_client.invoices.list()
        assert_matches_type(AsyncPage[Invoice], invoice, path=["response"])

    @parametrize
    async def test_method_list_with_all_params(self, async_client: AsyncOrb) -> None:
        invoice = await async_client.invoices.list(
            amount="amount",
            amount_gt="amount[gt]",
            amount_lt="amount[lt]",
            cursor="cursor",
            customer_id="customer_id",
            date_type="due_date",
            due_date=parse_date("2019-12-27"),
            due_date_window="due_date_window",
            due_date_gt=parse_date("2019-12-27"),
            due_date_lt=parse_date("2019-12-27"),
            external_customer_id="external_customer_id",
            invoice_date_gt=parse_datetime("2019-12-27T18:11:19.117Z"),
            invoice_date_gte=parse_datetime("2019-12-27T18:11:19.117Z"),
            invoice_date_lt=parse_datetime("2019-12-27T18:11:19.117Z"),
            invoice_date_lte=parse_datetime("2019-12-27T18:11:19.117Z"),
            is_recurring=True,
            limit=1,
            status=["draft"],
            subscription_id="subscription_id",
        )
        assert_matches_type(AsyncPage[Invoice], invoice, path=["response"])

    @parametrize
    async def test_raw_response_list(self, async_client: AsyncOrb) -> None:
        response = await async_client.invoices.with_raw_response.list()

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        invoice = response.parse()
        assert_matches_type(AsyncPage[Invoice], invoice, path=["response"])

    @parametrize
    async def test_streaming_response_list(self, async_client: AsyncOrb) -> None:
        async with async_client.invoices.with_streaming_response.list() as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            invoice = await response.parse()
            assert_matches_type(AsyncPage[Invoice], invoice, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    async def test_method_fetch(self, async_client: AsyncOrb) -> None:
        invoice = await async_client.invoices.fetch(
            "invoice_id",
        )
        assert_matches_type(Invoice, invoice, path=["response"])

    @parametrize
    async def test_raw_response_fetch(self, async_client: AsyncOrb) -> None:
        response = await async_client.invoices.with_raw_response.fetch(
            "invoice_id",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        invoice = response.parse()
        assert_matches_type(Invoice, invoice, path=["response"])

    @parametrize
    async def test_streaming_response_fetch(self, async_client: AsyncOrb) -> None:
        async with async_client.invoices.with_streaming_response.fetch(
            "invoice_id",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            invoice = await response.parse()
            assert_matches_type(Invoice, invoice, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    async def test_path_params_fetch(self, async_client: AsyncOrb) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `invoice_id` but received ''"):
            await async_client.invoices.with_raw_response.fetch(
                "",
            )

    @parametrize
    async def test_method_fetch_upcoming(self, async_client: AsyncOrb) -> None:
        invoice = await async_client.invoices.fetch_upcoming(
            subscription_id="subscription_id",
        )
        assert_matches_type(InvoiceFetchUpcomingResponse, invoice, path=["response"])

    @parametrize
    async def test_raw_response_fetch_upcoming(self, async_client: AsyncOrb) -> None:
        response = await async_client.invoices.with_raw_response.fetch_upcoming(
            subscription_id="subscription_id",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        invoice = response.parse()
        assert_matches_type(InvoiceFetchUpcomingResponse, invoice, path=["response"])

    @parametrize
    async def test_streaming_response_fetch_upcoming(self, async_client: AsyncOrb) -> None:
        async with async_client.invoices.with_streaming_response.fetch_upcoming(
            subscription_id="subscription_id",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            invoice = await response.parse()
            assert_matches_type(InvoiceFetchUpcomingResponse, invoice, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    async def test_method_issue(self, async_client: AsyncOrb) -> None:
        invoice = await async_client.invoices.issue(
            invoice_id="invoice_id",
        )
        assert_matches_type(Invoice, invoice, path=["response"])

    @parametrize
    async def test_method_issue_with_all_params(self, async_client: AsyncOrb) -> None:
        invoice = await async_client.invoices.issue(
            invoice_id="invoice_id",
            synchronous=True,
        )
        assert_matches_type(Invoice, invoice, path=["response"])

    @parametrize
    async def test_raw_response_issue(self, async_client: AsyncOrb) -> None:
        response = await async_client.invoices.with_raw_response.issue(
            invoice_id="invoice_id",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        invoice = response.parse()
        assert_matches_type(Invoice, invoice, path=["response"])

    @parametrize
    async def test_streaming_response_issue(self, async_client: AsyncOrb) -> None:
        async with async_client.invoices.with_streaming_response.issue(
            invoice_id="invoice_id",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            invoice = await response.parse()
            assert_matches_type(Invoice, invoice, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    async def test_path_params_issue(self, async_client: AsyncOrb) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `invoice_id` but received ''"):
            await async_client.invoices.with_raw_response.issue(
                invoice_id="",
            )

    @parametrize
    async def test_method_mark_paid(self, async_client: AsyncOrb) -> None:
        invoice = await async_client.invoices.mark_paid(
            invoice_id="invoice_id",
            payment_received_date=parse_date("2023-09-22"),
        )
        assert_matches_type(Invoice, invoice, path=["response"])

    @parametrize
    async def test_method_mark_paid_with_all_params(self, async_client: AsyncOrb) -> None:
        invoice = await async_client.invoices.mark_paid(
            invoice_id="invoice_id",
            payment_received_date=parse_date("2023-09-22"),
            external_id="external_payment_id_123",
            notes="notes",
        )
        assert_matches_type(Invoice, invoice, path=["response"])

    @parametrize
    async def test_raw_response_mark_paid(self, async_client: AsyncOrb) -> None:
        response = await async_client.invoices.with_raw_response.mark_paid(
            invoice_id="invoice_id",
            payment_received_date=parse_date("2023-09-22"),
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        invoice = response.parse()
        assert_matches_type(Invoice, invoice, path=["response"])

    @parametrize
    async def test_streaming_response_mark_paid(self, async_client: AsyncOrb) -> None:
        async with async_client.invoices.with_streaming_response.mark_paid(
            invoice_id="invoice_id",
            payment_received_date=parse_date("2023-09-22"),
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            invoice = await response.parse()
            assert_matches_type(Invoice, invoice, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    async def test_path_params_mark_paid(self, async_client: AsyncOrb) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `invoice_id` but received ''"):
            await async_client.invoices.with_raw_response.mark_paid(
                invoice_id="",
                payment_received_date=parse_date("2023-09-22"),
            )

    @parametrize
    async def test_method_pay(self, async_client: AsyncOrb) -> None:
        invoice = await async_client.invoices.pay(
            "invoice_id",
        )
        assert_matches_type(Invoice, invoice, path=["response"])

    @parametrize
    async def test_raw_response_pay(self, async_client: AsyncOrb) -> None:
        response = await async_client.invoices.with_raw_response.pay(
            "invoice_id",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        invoice = response.parse()
        assert_matches_type(Invoice, invoice, path=["response"])

    @parametrize
    async def test_streaming_response_pay(self, async_client: AsyncOrb) -> None:
        async with async_client.invoices.with_streaming_response.pay(
            "invoice_id",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            invoice = await response.parse()
            assert_matches_type(Invoice, invoice, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    async def test_path_params_pay(self, async_client: AsyncOrb) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `invoice_id` but received ''"):
            await async_client.invoices.with_raw_response.pay(
                "",
            )

    @parametrize
    async def test_method_void(self, async_client: AsyncOrb) -> None:
        invoice = await async_client.invoices.void(
            "invoice_id",
        )
        assert_matches_type(Invoice, invoice, path=["response"])

    @parametrize
    async def test_raw_response_void(self, async_client: AsyncOrb) -> None:
        response = await async_client.invoices.with_raw_response.void(
            "invoice_id",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        invoice = response.parse()
        assert_matches_type(Invoice, invoice, path=["response"])

    @parametrize
    async def test_streaming_response_void(self, async_client: AsyncOrb) -> None:
        async with async_client.invoices.with_streaming_response.void(
            "invoice_id",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            invoice = await response.parse()
            assert_matches_type(Invoice, invoice, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    async def test_path_params_void(self, async_client: AsyncOrb) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `invoice_id` but received ''"):
            await async_client.invoices.with_raw_response.void(
                "",
            )



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import List

from .tier import Tier
from ..._models import BaseModel

__all__ = ["TieredConfig"]


class TieredConfig(BaseModel):
    tiers: List[Tier]
    """Tiers for rating based on total usage quantities into the specified tier"""



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

import os
from typing import Any, cast

import pytest

from orb import Orb, AsyncOrb
from orb.types import (
    SubscriptionChangeApplyResponse,
    SubscriptionChangeCancelResponse,
    SubscriptionChangeRetrieveResponse,
)
from tests.utils import assert_matches_type

base_url = os.environ.get("TEST_API_BASE_URL", "http://127.0.0.1:4010")


class TestSubscriptionChanges:
    parametrize = pytest.mark.parametrize("client", [False, True], indirect=True, ids=["loose", "strict"])

    @parametrize
    def test_method_retrieve(self, client: Orb) -> None:
        subscription_change = client.subscription_changes.retrieve(
            "subscription_change_id",
        )
        assert_matches_type(SubscriptionChangeRetrieveResponse, subscription_change, path=["response"])

    @parametrize
    def test_raw_response_retrieve(self, client: Orb) -> None:
        response = client.subscription_changes.with_raw_response.retrieve(
            "subscription_change_id",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        subscription_change = response.parse()
        assert_matches_type(SubscriptionChangeRetrieveResponse, subscription_change, path=["response"])

    @parametrize
    def test_streaming_response_retrieve(self, client: Orb) -> None:
        with client.subscription_changes.with_streaming_response.retrieve(
            "subscription_change_id",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            subscription_change = response.parse()
            assert_matches_type(SubscriptionChangeRetrieveResponse, subscription_change, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    def test_path_params_retrieve(self, client: Orb) -> None:
        with pytest.raises(
            ValueError, match=r"Expected a non-empty value for `subscription_change_id` but received ''"
        ):
            client.subscription_changes.with_raw_response.retrieve(
                "",
            )

    @parametrize
    def test_method_apply(self, client: Orb) -> None:
        subscription_change = client.subscription_changes.apply(
            subscription_change_id="subscription_change_id",
        )
        assert_matches_type(SubscriptionChangeApplyResponse, subscription_change, path=["response"])

    @parametrize
    def test_method_apply_with_all_params(self, client: Orb) -> None:
        subscription_change = client.subscription_changes.apply(
            subscription_change_id="subscription_change_id",
            description="description",
            previously_collected_amount="previously_collected_amount",
        )
        assert_matches_type(SubscriptionChangeApplyResponse, subscription_change, path=["response"])

    @parametrize
    def test_raw_response_apply(self, client: Orb) -> None:
        response = client.subscription_changes.with_raw_response.apply(
            subscription_change_id="subscription_change_id",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        subscription_change = response.parse()
        assert_matches_type(SubscriptionChangeApplyResponse, subscription_change, path=["response"])

    @parametrize
    def test_streaming_response_apply(self, client: Orb) -> None:
        with client.subscription_changes.with_streaming_response.apply(
            subscription_change_id="subscription_change_id",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            subscription_change = response.parse()
            assert_matches_type(SubscriptionChangeApplyResponse, subscription_change, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    def test_path_params_apply(self, client: Orb) -> None:
        with pytest.raises(
            ValueError, match=r"Expected a non-empty value for `subscription_change_id` but received ''"
        ):
            client.subscription_changes.with_raw_response.apply(
                subscription_change_id="",
            )

    @parametrize
    def test_method_cancel(self, client: Orb) -> None:
        subscription_change = client.subscription_changes.cancel(
            "subscription_change_id",
        )
        assert_matches_type(SubscriptionChangeCancelResponse, subscription_change, path=["response"])

    @parametrize
    def test_raw_response_cancel(self, client: Orb) -> None:
        response = client.subscription_changes.with_raw_response.cancel(
            "subscription_change_id",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        subscription_change = response.parse()
        assert_matches_type(SubscriptionChangeCancelResponse, subscription_change, path=["response"])

    @parametrize
    def test_streaming_response_cancel(self, client: Orb) -> None:
        with client.subscription_changes.with_streaming_response.cancel(
            "subscription_change_id",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            subscription_change = response.parse()
            assert_matches_type(SubscriptionChangeCancelResponse, subscription_change, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    def test_path_params_cancel(self, client: Orb) -> None:
        with pytest.raises(
            ValueError, match=r"Expected a non-empty value for `subscription_change_id` but received ''"
        ):
            client.subscription_changes.with_raw_response.cancel(
                "",
            )


class TestAsyncSubscriptionChanges:
    parametrize = pytest.mark.parametrize(
        "async_client", [False, True, {"http_client": "aiohttp"}], indirect=True, ids=["loose", "strict", "aiohttp"]
    )

    @parametrize
    async def test_method_retrieve(self, async_client: AsyncOrb) -> None:
        subscription_change = await async_client.subscription_changes.retrieve(
            "subscription_change_id",
        )
        assert_matches_type(SubscriptionChangeRetrieveResponse, subscription_change, path=["response"])

    @parametrize
    async def test_raw_response_retrieve(self, async_client: AsyncOrb) -> None:
        response = await async_client.subscription_changes.with_raw_response.retrieve(
            "subscription_change_id",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        subscription_change = response.parse()
        assert_matches_type(SubscriptionChangeRetrieveResponse, subscription_change, path=["response"])

    @parametrize
    async def test_streaming_response_retrieve(self, async_client: AsyncOrb) -> None:
        async with async_client.subscription_changes.with_streaming_response.retrieve(
            "subscription_change_id",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            subscription_change = await response.parse()
            assert_matches_type(SubscriptionChangeRetrieveResponse, subscription_change, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    async def test_path_params_retrieve(self, async_client: AsyncOrb) -> None:
        with pytest.raises(
            ValueError, match=r"Expected a non-empty value for `subscription_change_id` but received ''"
        ):
            await async_client.subscription_changes.with_raw_response.retrieve(
                "",
            )

    @parametrize
    async def test_method_apply(self, async_client: AsyncOrb) -> None:
        subscription_change = await async_client.subscription_changes.apply(
            subscription_change_id="subscription_change_id",
        )
        assert_matches_type(SubscriptionChangeApplyResponse, subscription_change, path=["response"])

    @parametrize
    async def test_method_apply_with_all_params(self, async_client: AsyncOrb) -> None:
        subscription_change = await async_client.subscription_changes.apply(
            subscription_change_id="subscription_change_id",
            description="description",
            previously_collected_amount="previously_collected_amount",
        )
        assert_matches_type(SubscriptionChangeApplyResponse, subscription_change, path=["response"])

    @parametrize
    async def test_raw_response_apply(self, async_client: AsyncOrb) -> None:
        response = await async_client.subscription_changes.with_raw_response.apply(
            subscription_change_id="subscription_change_id",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        subscription_change = response.parse()
        assert_matches_type(SubscriptionChangeApplyResponse, subscription_change, path=["response"])

    @parametrize
    async def test_streaming_response_apply(self, async_client: AsyncOrb) -> None:
        async with async_client.subscription_changes.with_streaming_response.apply(
            subscription_change_id="subscription_change_id",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            subscription_change = await response.parse()
            assert_matches_type(SubscriptionChangeApplyResponse, subscription_change, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    async def test_path_params_apply(self, async_client: AsyncOrb) -> None:
        with pytest.raises(
            ValueError, match=r"Expected a non-empty value for `subscription_change_id` but received ''"
        ):
            await async_client.subscription_changes.with_raw_response.apply(
                subscription_change_id="",
            )

    @parametrize
    async def test_method_cancel(self, async_client: AsyncOrb) -> None:
        subscription_change = await async_client.subscription_changes.cancel(
            "subscription_change_id",
        )
        assert_matches_type(SubscriptionChangeCancelResponse, subscription_change, path=["response"])

    @parametrize
    async def test_raw_response_cancel(self, async_client: AsyncOrb) -> None:
        response = await async_client.subscription_changes.with_raw_response.cancel(
            "subscription_change_id",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        subscription_change = response.parse()
        assert_matches_type(SubscriptionChangeCancelResponse, subscription_change, path=["response"])

    @parametrize
    async def test_streaming_response_cancel(self, async_client: AsyncOrb) -> None:
        async with async_client.subscription_changes.with_streaming_response.cancel(
            "subscription_change_id",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            subscription_change = await response.parse()
            assert_matches_type(SubscriptionChangeCancelResponse, subscription_change, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    async def test_path_params_cancel(self, async_client: AsyncOrb) -> None:
        with pytest.raises(
            ValueError, match=r"Expected a non-empty value for `subscription_change_id` but received ''"
        ):
            await async_client.subscription_changes.with_raw_response.cancel(
                "",
            )



# File generated from our OpenAPI spec by Stainless.

from __future__ import annotations

import os
from typing import Any, cast
from datetime import datetime, timedelta

import pytest
import time_machine

from orb import Orb, AsyncOrb

base_url = os.environ.get("TEST_API_BASE_URL", "http://127.0.0.1:4010")


class TestWebhooks:
    parametrize = pytest.mark.parametrize("client", [False, True], indirect=True, ids=["loose", "strict"])

    timestamp = "2024-03-27T15:42:29.551"
    fake_now = datetime.fromisoformat(timestamp).astimezone()

    payload = """{"id": "o4mmewpfNNTnjfZc", "created_at": "2024-03-27T15:42:29+00:00", "type": "resource_event.test", "properties": {"message": "A test webhook from Orb. Happy testing!"}}"""
    signature = "9d25de966891ab0bc18754faf8d83d0980b44ae330fcc130b41a6cf3daf1f391"
    headers = {
        "X-Orb-Timestamp": timestamp,
        "X-Orb-Signature": f"v1={signature}",
    }
    secret = "c-UGKYdnhHh436B_sMouYAPUvXyWpzOdunZBV5dFSD8"

    @time_machine.travel(fake_now)
    def test_unwrap(self, client: Orb) -> None:
        payload = self.payload
        headers = self.headers
        secret = self.secret

        client.webhooks.unwrap(payload, headers, secret=secret)

    @time_machine.travel(fake_now)
    def test_verify_signature(self, client: Orb) -> None:
        payload = self.payload
        headers = self.headers
        secret = self.secret
        signature = self.signature
        verify = client.webhooks.verify_signature

        assert verify(payload=payload, headers=headers, secret=secret) is None

        with pytest.raises(ValueError, match="Webhook timestamp is too old"):
            with time_machine.travel(self.fake_now + timedelta(minutes=6)):
                assert verify(payload=payload, headers=headers, secret=secret) is False

        with pytest.raises(ValueError, match="Webhook timestamp is too new"):
            with time_machine.travel(self.fake_now - timedelta(minutes=6)):
                assert verify(payload=payload, headers=headers, secret=secret) is False

        invalid_signature_message = "None of the given webhook signatures match the expected signature"
        with pytest.raises(ValueError, match=invalid_signature_message):
            verify(
                payload=payload,
                headers=headers,
                secret="foo",
            )

        # multiple signatures
        invalid_signature = "my_invalid_signature"
        assert (
            verify(
                payload=payload,
                headers={**headers, "X-Orb-Signature": f"v1={invalid_signature} v1={signature}"},
                secret=secret,
            )
            is None
        )

        # different signature version
        with pytest.raises(ValueError, match=invalid_signature_message):
            verify(
                payload=payload,
                headers={**headers, "X-Orb-Signature": f"v2={signature}"},
                secret=secret,
            )

        assert (
            verify(
                payload=payload,
                headers={**headers, "X-Orb-Signature": f"v2={signature} v1={signature}"},
                secret=secret,
            )
            is None
        )

        # missing version
        with pytest.raises(ValueError, match=invalid_signature_message):
            verify(
                payload=payload,
                headers={**headers, "X-Orb-Signature": signature},
                secret=secret,
            )

        # non-string payload
        with pytest.raises(ValueError, match=r"Webhook body should be a string"):
            verify(
                payload=cast(Any, {"payload": payload}),
                headers=headers,
                secret=secret,
            )


class TestAsyncWebhooks:
    parametrize = pytest.mark.parametrize("async_client", [False, True], indirect=True, ids=["loose", "strict"])

    timestamp = "2024-03-27T15:42:29.551"
    fake_now = datetime.fromisoformat(timestamp).astimezone()

    payload = """{"id": "o4mmewpfNNTnjfZc", "created_at": "2024-03-27T15:42:29+00:00", "type": "resource_event.test", "properties": {"message": "A test webhook from Orb. Happy testing!"}}"""
    signature = "9d25de966891ab0bc18754faf8d83d0980b44ae330fcc130b41a6cf3daf1f391"
    headers = {
        "X-Orb-Timestamp": timestamp,
        "X-Orb-Signature": f"v1={signature}",
    }
    secret = "c-UGKYdnhHh436B_sMouYAPUvXyWpzOdunZBV5dFSD8"

    @time_machine.travel(fake_now)
    def test_unwrap(self, async_client: AsyncOrb) -> None:
        payload = self.payload
        headers = self.headers
        secret = self.secret

        async_client.webhooks.unwrap(payload, headers, secret=secret)

    @time_machine.travel(fake_now)
    def test_verify_signature(self, async_client: Orb) -> None:
        payload = self.payload
        headers = self.headers
        secret = self.secret
        signature = self.signature
        verify = async_client.webhooks.verify_signature

        assert verify(payload=payload, headers=headers, secret=secret) is None

        with pytest.raises(ValueError, match="Webhook timestamp is too old"):
            with time_machine.travel(self.fake_now + timedelta(minutes=6)):
                assert verify(payload=payload, headers=headers, secret=secret) is False

        with pytest.raises(ValueError, match="Webhook timestamp is too new"):
            with time_machine.travel(self.fake_now - timedelta(minutes=6)):
                assert verify(payload=payload, headers=headers, secret=secret) is False

        invalid_signature_message = "None of the given webhook signatures match the expected signature"
        with pytest.raises(ValueError, match=invalid_signature_message):
            verify(
                payload=payload,
                headers=headers,
                secret="foo",
            )

        # multiple signatures
        invalid_signature = "my_invalid_signature"
        assert (
            verify(
                payload=payload,
                headers={**headers, "X-Orb-Signature": f"v1={invalid_signature} v1={signature}"},
                secret=secret,
            )
            is None
        )

        # different signature version
        with pytest.raises(ValueError, match=invalid_signature_message):
            verify(
                payload=payload,
                headers={**headers, "X-Orb-Signature": f"v2={signature}"},
                secret=secret,
            )

        assert (
            verify(
                payload=payload,
                headers={**headers, "X-Orb-Signature": f"v2={signature} v1={signature}"},
                secret=secret,
            )
            is None
        )

        # missing version
        with pytest.raises(ValueError, match=invalid_signature_message):
            verify(
                payload=payload,
                headers={**headers, "X-Orb-Signature": signature},
                secret=secret,
            )

        # non-string payload
        with pytest.raises(ValueError, match=r"Webhook body should be a string"):
            verify(
                payload=cast(Any, {"payload": payload}),
                headers=headers,
                secret=secret,
            )



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import Dict, Optional
from typing_extensions import Annotated, TypedDict

from ..._utils import PropertyInfo

__all__ = ["ExternalDimensionalPriceGroupIDUpdateParams"]


class ExternalDimensionalPriceGroupIDUpdateParams(TypedDict, total=False):
    body_external_dimensional_price_group_id: Annotated[
        Optional[str], PropertyInfo(alias="external_dimensional_price_group_id")
    ]
    """
    An optional user-defined ID for this dimensional price group resource, used
    throughout the system as an alias for this dimensional price group. Use this
    field to identify a dimensional price group by an existing identifier in your
    system.
    """

    metadata: Optional[Dict[str, Optional[str]]]
    """User-specified key/value pairs for the resource.

    Individual keys can be removed by setting the value to `null`, and the entire
    metadata mapping can be cleared by setting `metadata` to `null`.
    """
